import subprocess
import cPickle
import json
import os

import ajenti
from ajenti.api import *
from ajenti.api.http import HttpPlugin, url
from ajenti.plugins.main.api import SectionPlugin
from ajenti.plugins.services.api import ServiceMultiplexor
from ajenti.ui import on
from ajenti.ui.binder import Binder
from ajenti.plugins import manager

from reconfigure.configs import SambaConfig
from reconfigure.items.samba import ShareData


SHARE_NAME = '__projects'
ROOT = '/var/lib/ajenti/elements-projects'


class ElementsProject (object):
    def __init__(self):
        self.name = ''
        self.path = '/'
        self.permissions = []
        self.lock = None

    def verify(self):
        if not hasattr(self, 'lock'):
            self.lock = None

    def json(self):
        return {
            'name': self.name,
            'path': self.path,
            'lock': self.lock,
        }

    def is_allowed_for(self, user):
        for p in self.permissions:
            if p.user == user:
                return True
        return False


class ElementsProjectPermission (object):
    def __init__(self):
        self.project = None
        self.user = 'root'


@plugin
class ElementsProjectManager (BasePlugin):
    default_classconfig = {'projects': ''}
    classconfig_root = True

    def init(self):
        self.smb_config = SambaConfig(path='/etc/samba/smb.conf')

        try:
            self.projects = cPickle.loads(str(self.classconfig['projects']))
        except:
            self.projects = []

        for p in self.projects:
            p.verify()

    def get_project(self, name):
        for p in self.projects:
            if p.name == name:
                return p
        return None

    def lock(self, name, user):
        p = self.get_project(name)
        if p.lock and p.lock != user:
            return False
        p.lock = user
        return True

    def unlock(self, name, user):
        p = self.get_project(name)
        if not p.lock or p.lock == user:
            p.lock = None
            return True
        return False

    def save(self):
        self.smb_config.load()

        for share in self.smb_config.tree.shares:
            if share.name == SHARE_NAME:
                self.smb_config.tree.shares.remove(share)

        s = ShareData()
        s.name = SHARE_NAME
        s.path = ROOT
        s.browseable = False
        s.read_only = False
        s.guest_ok = True
        self.smb_config.tree.shares.append(s)
        self.smb_config.save()

        if not os.path.exists(ROOT):
            os.mkdir(ROOT)

        for f in os.listdir(ROOT):
            path = os.path.join(ROOT, f)
            subprocess.call(['umount', '-f', path])
            os.rmdir(path)

        for p in self.projects:
            path = os.path.join(ROOT, p.name)
            os.mkdir(path)
            try:
                subprocess.check_output(['mount', p.path, path, '-o', 'bind'])
            except:
                print 'Couldn\'t mount project %s' % p.name

        ServiceMultiplexor.get().get_one('smb' if ajenti.platform == 'centos' else 'smbd').restart()

        self.classconfig['projects'] = cPickle.dumps(self.projects)
        self.save_classconfig()


@plugin
class ElementsProjects (SectionPlugin):
    def init(self):
        self.title = 'Projects'
        self.icon = 'suitcase'
        self.category = 'Elements'
        self.append(self.ui.inflate('elements:projects-main'))

        #def post_graph_bind(o, c, i, u):
        #    for plot in u.nearest(lambda x: x.typeid == 'munin:plot'):
        #        plot.on('widget', self.on_add_widget, i)
        #
        #self.find('graphs').post_item_bind = post_graph_bind

        self.find('projects').new_item = lambda c: ElementsProject()
        self.find('permissions').new_item = lambda c: ElementsProjectPermission()
        self.find('user-dropdown').labels = self.find('user-dropdown').values = list(ajenti.config.tree.users)

        self.mgr = ElementsProjectManager.get(manager.context)
        self.binder = Binder(None, self)

    def on_page_load(self):
        self.binder.reset(self.mgr).autodiscover().populate()

    @on('save', 'click')
    def save(self):
        self.binder.update()
        self.mgr.save()


@plugin
class ElementsProjectServer (HttpPlugin):
    def init(self):
        self.mgr = ElementsProjectManager.get(manager.context)

    @url('/elements/projects/(?P<endpoint>.+)')
    def get_page(self, context, endpoint):
        if context.session.identity is None:
            context.respond_forbidden()

        context.respond_ok()

        endpoint = endpoint.split('/')

        if endpoint[0] == 'test':
            return 'ok'

        if endpoint[0] == 'lock-project':
            self.mgr.lock(endpoint[1], context.session.identity)
            return 'ok'

        if endpoint[0] == 'unlock-project':
            self.mgr.unlock(endpoint[1], context.session.identity)
            return 'ok'

        if endpoint[0] == 'unlock-all':
            for p in self.mgr.projects:
                self.mgr.unlock(p.name, context.session.identity)
            return 'ok'

        if endpoint[0] == 'list-projects':
            return json.dumps([p.json() for p in self.mgr.projects if p.is_allowed_for(context.session.identity)])
